// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	hkgtigen.cc --- HK parameter driven Good-Time-Interval GENeration
//
//	DESCRIPTION:
//		tabgtigen is used to create Good-Time-Interval files from
//		tabular data in a data set driven by a user-specified boolean
//		expression. The latter contains names of table columns
//		in the data set as variables.
//
//	CHANGE HISTORY:
//		UL, 1998-10-20: 0.1.0: - initial version; bases on draft task
//								 description v0.1
//		UL, 1999-10-21: 0.2.0: - a few internal tweaks
//		UL, 1999-10-22: 0.3.0: - added warning about empty GTI list
//		UL, 2001-07-24: 1.0.0: - S/C HK parameters were not properly taken
//								 into account
//							   - consistent use of namespace std
//		UL, 2003-03-14: 2.0.0: - significant changes related to implementation
//								 of SCR 108: add diagnostic mode
//
//	AUTHOR:
//		U. Lammers, Tue Oct 19 16:27:15 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Expanded task description of hkgtigen
//
//	$Header: /home/sasbuild/repositories/sasdev/task_devel/packages/hkgtigen/src/hkgtigen.cc,v 1.12 2003/04/08 08:58:27 gvacanti Exp $
//
#include <algorithm>	//	find
#include <fstream>
#include <iomanip>

#include "param.h"
#include "errstr.h"
#include "stringutil.h"
#include "Xmm.h"
#include "HKvalRangesServer.h"	//	CAL
#include "CalServer.h"
#include "OalServer.h"			//	OAL
#include "GtiTableFilter.h"		//	tabgtigen
#include "dsplot.h"

#include "hkgtigen.h"

using namespace std;

class GtiHKvalRanges : public HKvalRanges {
public:
	GtiHKvalRanges(Xmm::Instrument instrument = Xmm::EMOS1)
	{
		//
		//	Instantiate GtiHkvalRanges object - retrieve all entries from
		//	CAL for instrument <instrument>
		//
		calServer.setInstrument(instrument);
		HKvalRangesServer *p = 0; p = calServer.getAtom(p);
		HKvalRanges &r = *this;
		r = *p;
		calServer.releaseAtom(p);
	} // CalHKvalRanges::CalHKvalRanges()


	void allRangesOfParameters(const std::vector<std::string> &parNames,
							   bool except = false)
	{
 		for (HKvalRanges::MapType::iterator i = begin();i != end();) {
			HKvalRanges::MapType::iterator current = i++;
			std::vector<std::string>::const_iterator j =
				std::find(parNames.begin(),parNames.end(),(*current).first);

			if ((!except && j != parNames.end()) ||
				except && j == parNames.end())
				{
				    continue;
				}
			erase(current);
		}
	} // CalHKvalRanges::allRangesOfParameters()


	void augmentRanges(const std::vector<std::string> &parNames,
					   const std::vector<std::string> &parRanges)
	{
		if (parNames.size() != parRanges.size())
			errstr<<"There are "<<parNames.size()<<" parameter names for "
					"range augmentation but only "<<parRanges.size()<<" ranges"
				  <<Sas::error("SizeMismatch");

		for (unsigned n=0; n<parNames.size(); ++n)
			(*this)[parNames[n]] = parRanges[n];
	} // CalHKvalRanges::augmentRanges()


	~GtiHKvalRanges() { }
};


class StringVector : public std::vector<std::string> {
public:
	explicit StringVector(std::vector<std::string> &v) : _v(v) {}
	void operator()(std::string &s) { s = uppercase(s); }

private:
	std::vector<std::string> &_v;
};

namespace {
	//
	//	get the names of all columns in the first table of the ODF
	//	data set for instrument/dataMode/eposureNr tuple
	//	<inst>/<dataMode>/1 - return column names as vector of strings
	//
	std::vector<std::string> hkColNames(Xmm::Instrument inst,
										Xmm::DataMode dataMode)
	{
		std::vector<std::string> l;

	    const OdfSet *odfSet =
				oalServer.odf()->selectSet(inst, dataMode, 1);
		DataSet *set = dataSetServer->open(odfSet->name(), Dal::Read,
										   DataSetServer::Low);
		Table *tab = set->table(0);
		unsigned nCols = tab->columns();
		for (unsigned n=0; n<nCols; l.push_back(tab->column(n++)->name()));
		dataSetServer->close(set);

		return l;
	} // hkColNames()

	//
	//	issue diagnostic message
	//
	void diagnostic (const std::string &name, const std::string &expr,
					 const std::string &timeCol, double pref, double postf)
	{
		errstr<<"creating GTI list from table="<<name<<" expression="<<expr
			  <<" timecolumn="<<timeCol<<" prefraction="<<pref
			  <<" postfraction="<<postf
			  <<Sas::message(Msg::App, Msg::Verbose);
	} // diagnostic()
}

void hkgtigen()
{
	//
	//	inquire instrument and obsid
	//
	Xmm::Instrument inst = Xmm::toInstrument(stringParameter("instrument"));
	stringstream s;
	s<<setw(4)<<setfill('0')<<oalServer.odf()->orbitNr()<<'_'
	 <<oalServer.odf()->proposalId()<<oalServer.odf()->observationId();
	string obsId(s.str());

	//
	//	get validity ranges from CAL and modify according to parameters
	//
	GtiHKvalRanges	valRanges(inst);
	if (booleanParameter("withparameters")) {
		std::vector<std::string> p = stringListParameter("parameters");
		for_each(p.begin(), p.end(), StringVector(p));
		valRanges.allRangesOfParameters(p, booleanParameter("except"));
	}
	
	//
	//	consider override parameters
	//
	if (booleanParameter("withoverrideparameters")) {
		std::vector<std::string> p = stringListParameter("overrideparameters");
		for_each(p.begin(), p.end(), StringVector(p));
		std::vector<std::string>
				vals(breakup(stringParameter("overrideparametervalues"), ";"));
		valRanges.augmentRanges(p, vals);
	}

	std::vector<Xmm::DataMode> hkkinds =
								xmmPayLoad.instrument(inst).periodicHKmodes();
	unsigned N = hkkinds.size();
	std::vector<Xmm::Instrument> hkinst(N, inst);

	//
	//	... add S/C HK modes
	//
	hkkinds.insert(hkkinds.end(), xmmPayLoad.xmm.periodicHKmodes().begin(),
								  xmmPayLoad.xmm.periodicHKmodes().end());
	std::fill_n(std::back_inserter(hkinst), hkkinds.size()-N, Xmm::SC);

	bool withGtiSet = booleanParameter("withgtiset"),
		 diagnosticMode = booleanParameter("diagnosticmode");

	//
	//	gtiList[0] will contain the or-result of all individual HK GTI lists
	//	gtiList[i] (i>=1) shall contin the GTIs related to HK parameter i
	//
	vector<GtiList *> gtiList;
	gtiList.push_back(new GtiList(true)); // intial GTI list - all times good

	vector<pair<string, string> > hkParms;// names and expressions; first
										  // elements contains total GTI
	
	// hkParms.push_back(make_pair("TOTAL", ""));
	// gcc-2.95.3 wants an explicit string
	hkParms.push_back(make_pair(string("TOTAL"), string("")));
	vector<pair<unsigned, vector<string> > > hktypes;
	double t0 = 1.e20,	//	t0: earliest time seen in HK
		   t1 = 0.;		//	t1: latest time seen in HK
	unsigned fromi = 1;

	//
	//	loop over HK parameter types
	//
	for (unsigned n=0; n<hkkinds.size(); ++n) {
		unsigned nExps = oalServer.odf()->exposures(hkinst[n], hkkinds[n]);
		if (!nExps)
			continue;

		const std::vector<std::string> colNames =
										hkColNames(hkinst[n], hkkinds[n]);
		HKvalRanges subSetRanges = valRanges.reduceWithShortNames(colNames);
		std::string expr = subSetRanges.selectionExpression();
		if (!expr.length())
			continue;

		vector<GtiTableFilter *> tableFilter;
		tableFilter.push_back(new GtiTableFilter(expr));
		if (diagnosticMode) {
			//	instantiate individual GtiTable filters
			for (HKvalRanges::MCIter i=subSetRanges.begin();
				 i!=subSetRanges.end(); ++i) {
				tableFilter.push_back(new GtiTableFilter(
												subSetRanges.getExpanded(i)));
				gtiList.push_back(new GtiList(true));
				hkParms.push_back(make_pair(i->first, i->second));
			}
		}

		//
		//	loop over the exposures - in the early days HK data were
		//	split by exposure
		//
		std::string timeColumn = stringParameter("timecolumn");
		double preFraction = realParameter("prefraction"),
			   postFraction = realParameter("postfraction");

		vector<string> hksets;
		for (unsigned i=1; i<=nExps; ++i) {
		    const OdfSet *hkSet =
					oalServer.odf()->selectSet(hkinst[n], hkkinds[n], i);
			hksets.push_back(Filename::split(hkSet->name()).second);

			diagnostic(hkSet->name(), expr, timeColumn, preFraction,
					   postFraction);

			if (withGtiSet) {
				//
				//	create the GTI list and append it to global list
				//
				diagnostic(hkSet->name(), tableFilter[0]->expression(),
						   timeColumn, preFraction, postFraction);
				
				GtiList l = tableFilter[0]->gtis(SetTableColumn(hkSet->name()),
												 timeColumn,
												 preFraction, postFraction);
				t0 = std::min(t0, tableFilter[0]->earliestTime());
				t1 = std::max(t1, tableFilter[0]->latestTime());

				*gtiList[0] &= l;

				if (diagnosticMode)
					for (unsigned k=1; k<tableFilter.size(); ++k) {
						diagnostic(hkSet->name(), tableFilter[k]->expression(),
								   timeColumn, preFraction, postFraction);
						GtiList l = tableFilter[k]->gtis(
												SetTableColumn(hkSet->name()),
												timeColumn,
												preFraction, postFraction);
						*gtiList[k+fromi-1] &= l;
					}
			}
		}
		for (unsigned k=0; k<tableFilter.size(); delete tableFilter[k++]);

		hktypes.push_back(make_pair(fromi, hksets));
		fromi = gtiList.size();
	}

	//
	//	by now all HK ranges should have been used - if not print warning
	//	message about remaning ones
	//
	if (valRanges.size()) {
		errstr<<"The following validity ranges are not considered in the "
				"GTI list generation:\n";
		for(HKvalRanges::MIter i = valRanges.begin(); i!= valRanges.end(); ++i)
			errstr<<(*i).first<<" : "<<(*i).second<<std::endl;
		errstr<<Sas::warning("UnusedValidityRanges");
	}

	//
	//	write the final GTI list to a data set and print out statistical
	//	information
	//
	ostream* os = &cout;
	if (diagnosticMode) {
		ofstream *ofs = new ofstream;
		string diagFile = stringParameter("diagfile");
		if (uppercase(diagFile) != "STDOUT") {
			ofs->open(diagFile.c_str());
			if (!*ofs)
				errstr<<"could not open "<<diagFile<<Sas::error("openFile");
			os = ofs;
		}
	}

	double tot = t1-t0;	//	total observation time determined from
						//	HK data
	if (withGtiSet) {
		std::string gtiSetName = stringParameter("gtiset"); {
		OutOGIPgtiTable outOGIPgtiTable(gtiSetName + ":STDGTI");
		outOGIPgtiTable<<*gtiList[0];
		if (diagnosticMode) {
			*os<<"#\n# HK parameter out-of-limit diagnostic file\n"
			     "# ODF: "<<obsId<<" - instrument: "<<Xmm::toString(inst)<<endl
			   <<"# 1. column: HK parameter name\n"
				 "# 2. column: fraction of observation parameter is out of "
				 "valid bounds\n"
				 "#\n";
			*os<<hkParms[0].first<<"\t\t"
			   <<std::min(std::max(1.-gtiList[0]->ontime()/tot, 0.), 1.)
			   <<endl;
			unsigned ind = 0;
			for (unsigned n=1; n<gtiList.size(); ++n) {
				OutOGIPgtiTable outtab(gtiSetName + ":STDGTI-" +
									   hkParms[n].first);
				outtab<<*gtiList[n];
				*outtab.table()->addAttribute("EXPR", "selection expression")
														= hkParms[n].second;
				if (ind<hktypes.size() && n>=hktypes[ind].first) {
					*os<<"# Parameters in set(s) ";
					for (unsigned k=0; k<hktypes[ind].second.size(); ++k)
						*os<<hktypes[ind].second[k]<<' ';
					*os<<endl;
					++ind;
				}
				*os<<hkParms[n].first<<"\t\t"
				   <<std::min(std::max(1.-gtiList[n]->ontime()/tot, 0.), 1.)
				   <<endl;
			}
		}
		} // OutOGIPGtiTable goes out of scope and is closed

		if (!gtiList[0]->size())
			errstr<<"The generated GTI data set `"<<gtiSetName<<"' is empty - "
					"applying it to any event list will result in no events "
					"getting selected."
				  <<Sas::warning("EmptyGtiSet");

		if (booleanParameter("plotgtis")) {
			//	visiualize the GTIs
			openParameters("dsplot");
			setParameter("withtable", false);
			setParameter("withoffsetx", true);
			setParameter("title", string("HK parameter GTIs - instrument ")
						 + Xmm::toString(inst));
			setParameter("subtitle", string("ODF ") + s.str());
			setParameter("withxranges", true);
			vector<string> tabs; tabs.push_back(gtiSetName);
			setParameter("xrangetables", tabs);
			setParameter("withalltablesafter", true);

			task(dsplot);
		}
	}

	//	clean up
	for (unsigned n=0; n<gtiList.size(); delete gtiList[n++]);
} // hkgtigen()
